import argparse
import inspect
import logging
import os
import sys
from pathlib import Path

from cpbox.tool import datatypes
from cpbox.tool import system

class BaseCliApp(object):

    def __init__(self):
        self.cli_root_dir = self._get_cli_root_dir()
        from cpbox.app.rootcontext import gcontext
        gcontext.try_enable_framework_mode()
        self.logger = logging.getLogger(self.__class__.__name__)

    def _get_cli_root_dir(self):
        if hasattr(sys.modules['__main__'], '__file__'):
            return Path(sys.argv[0]).resolve().parent
        else:
            return Path.cwd()

    def run_cmd_ret(self, cmd, log=True):
        return self.run_cmd(cmd, log=log)[1]

    def run_cmd(self, cmd, log=True):
        if log:
            self.logger.info('run_cmd: %s', cmd)
        return system.run_cmd(cmd)

    def shell_run(self, cmd, keep_pipeline=True, exit_on_error=True, dry_run=False, log=True):
        if log:
            self.logger.info('shell_run: %s', cmd)
        if dry_run:
            return 0
        return system.shell_run(cmd, keep_pipeline=keep_pipeline, exit_on_error=exit_on_error)

def append_cli_for_composition(container, cli):
    cli_map = {}
    if hasattr(container, '__cli_map__'):
        cli_map = container.__cli_map__
    for key, attr in cli.__class__.__dict__.items():
        if key[0:1] != '_' and callable(attr):
            cli_map[key] = cli
            setattr(container, key, attr.__get__(cli))

def merge_class_for_composition(*proxy_cls_list):
    def real_decorator(cls):
        cls.__proxy_src_cls__ = cls
        cls.__proxy_cls_list__ = list(proxy_cls_list)
        return cls
    return real_decorator

def keep_class_for_inheritance(cls):
    cls.keep_class_for_cli = True
    return cls

def keep_method_for_inheritance(f):
    f.keep_method_for_cli = True
    return f

class _MethodSpec(object):

    def __init__(self, func, base_method_spec=None):
        self.func = func
        self.method_name = func.__name__
        spec = inspect.getfullargspec(func)

        pos_args = [arg for arg in spec.args if arg != 'self']
        defaults = spec.defaults or ()
        default_values = dict(zip(spec.args[-len(defaults):], defaults)) if defaults else {}
        kwonlyargs = spec.kwonlyargs or []
        kwonlydefaults = spec.kwonlydefaults or {}
        varargs = spec.varargs

        self.named_args = pos_args + list(kwonlyargs)
        self.default_values = dict(default_values, **kwonlydefaults)
        self.varargs = varargs
        self.annotations = self._extract_annotations(spec)
        self.method_origin_args = pos_args + ([self.varargs] if self.varargs else []) + list(kwonlyargs)

        if base_method_spec is not None:
            self.named_args = list(dict.fromkeys(self.named_args + base_method_spec.named_args))
            self.default_values = dict(base_method_spec.default_values, **self.default_values)
            self.annotations = dict(base_method_spec.annotations, **self.annotations)
        self.parameters = self._build_parameters()

    def _extract_annotations(self, spec):
        # 注解值: 未开 PEP 563 时内置标量为类型对象 (int/str/bool/float);
        # 开了 from __future__ import annotations 则全为字符串
        source_annotations = spec.annotations or {}
        annotations = {}
        for arg in self.named_args:
            if arg in source_annotations:
                annotations[arg] = source_annotations[arg]
        if self.varargs and self.varargs in source_annotations:
            annotations[self.varargs] = source_annotations[self.varargs]
        return annotations

    def _build_parameters(self):
        parameters = {}
        for arg_name in self.named_args:
            annotation = self.annotations.get(arg_name)
            annotation_type_name = datatypes.resolve_scalar_type_name_from_annotation(annotation)
            if annotation is not None and annotation_type_name is None:
                raise ValueError('unsupported annotation for %s.%s: %s' % (self.method_name, arg_name, annotation))
            type_name = annotation_type_name
            if type_name is None and arg_name in self.default_values:
                type_name = datatypes.resolve_scalar_type_name_from_value(self.default_values[arg_name])
            if type_name is None:
                type_name = 'str'
            des = 'type: %s' % type_name
            if arg_name in self.default_values:
                des = '%s, default value is: %s' % (des, self.default_values[arg_name])
            parameters[arg_name] = {'type': type_name, 'des': des}
        return parameters

    def build_help_text(self):
        parameters = self.parameters or {}
        default_values = self.default_values or {}
        args = self.named_args or []
        varargs = self.varargs
        parts = []
        for arg in args:
            param = parameters.get(arg, {})
            arg_type = param.get('type', 'str')
            if len(arg) == 1:
                display_arg = '-' + arg
            else:
                display_arg = _build_main_long_option(arg)
            if arg in default_values:
                parts.append(f"{display_arg}[{arg_type}={default_values.get(arg)}]")
            else:
                parts.append(f"{display_arg}[{arg_type}*]")
        if varargs:
            parts.append(f"{varargs}... (remainder)")
        if not parts:
            return ''
        return ', '.join(parts)

class AppSpec(object):

    def __init__(self, app):
        self.base_method_spec = _MethodSpec(app.__init__)
        self.app = app
        self._method_processor = _MethodProcessor(app, self.base_method_spec)
        self._method_names_by_order = self._method_processor.get_method_names_by_order()
        self._methods_by_name = self._method_processor.get_methods_by_name()
        self._method_spec_map = self._method_processor.get_method_spec_map()
        self.method_will_call = None

    def get_method_names_by_order(self):
        return list(self._method_names_by_order)

    def get_methods_by_name(self):
        return dict(self._methods_by_name)

    def get_method_spec_map(self):
        return dict(self._method_spec_map)

    def fix_terminal_columns(self):
        if os.name == 'nt':
            return
        if sys.stdin.isatty() and sys.stdout.isatty():
            rows, columns = os.popen('stty size', 'r').read().split()
            os.environ['COLUMNS'] = columns

    def resolve(self, default_method=None):
        self.fix_terminal_columns()

        raw_method_name = None
        if len(sys.argv) > 1:
            raw_method_name = sys.argv[1]
        method_will_call = self._method_processor.resolve_method_name(raw_method_name, default_method)
        if not method_will_call:
            self.show_help_then_exit()

        self.method_will_call = method_will_call
        self._parse_args()

    def _parse_args(self):
        method_will_call = self.method_will_call
        method_spec = self._method_spec_map[method_will_call]
        kwargs = build_argparser_for_func(method_will_call, method_spec)

        self.base_kwargs = {}
        common_args_keys = set(self.base_method_spec.named_args)
        methods_args_keys = set(method_spec.method_origin_args)
        method_kwargs = {}
        for key, value in kwargs.items():
            if key in common_args_keys:
                self.base_kwargs[key] = value
            if key in methods_args_keys:
                method_kwargs[key] = value
        self.kwargs = method_kwargs

    def run(self):
        # kwargs 含 --a --b 等具名参数及 varargs 的 list（命令行剩余项）。
        # method(a,b,*rest) 需 a,b 作位置、*rest 作可变位置；method(*rest,a,b) 需 *rest 作可变位置、a,b 作 keyword-only。
        # 按 method_origin_args 顺序拆分：before_varargs 按位置传，varargs 按 * 展开，after_varargs 按关键字传。
        kwargs = dict(self.kwargs)
        method_spec = self._method_spec_map[self.method_will_call]
        # origin_args 如 [a,b,rest] 或 [rest,a,b]，保持 Python 签名顺序；varargs_name 如 'rest'、'files'，无则 None
        origin_args = method_spec.method_origin_args
        varargs_name = method_spec.varargs
        idx = origin_args.index(varargs_name) if varargs_name else len(origin_args)
        # before_varargs：varargs 前的参数，必须按位置传（如 method(a,b,*rest) 的 a,b）
        # after_varargs：varargs 后的参数，必须按关键字传（如 method(*rest,a,b) 的 a,b）
        before_varargs = origin_args[:idx]
        after_varargs = origin_args[idx + 1:] if varargs_name else []
        # positional_vals 为前 N 个位置参数；varargs_val 为 *args 的 tuple；keyword_kwargs 为 keyword-only 参数字典
        positional_vals = [kwargs.pop(x) for x in before_varargs]
        varargs_val = tuple(kwargs.pop(varargs_name, [])) if varargs_name else ()
        keyword_kwargs = {x: kwargs.pop(x) for x in after_varargs}
        _app = self.app(**self.base_kwargs)
        func = getattr(_app, self.method_will_call)
        # 按签名顺序调用：位置 + *varargs + keyword-only
        func(*positional_vals, *varargs_val, **keyword_kwargs)

    def show_help_then_exit(self):
        parser = argparse.ArgumentParser()
        parser.allow_abbrev = False
        subparsers = parser.add_subparsers(dest='method_name', metavar='cmd')
        for method_name in self._method_names_by_order:
            help_text = self._method_processor.build_help_for_method(method_name)
            subparsers.add_parser(_to_kebab(method_name), help=help_text, add_help=False)
        parser.print_help()
        if any(arg in ('-h', '--help') for arg in sys.argv[1:]):
            sys.exit(0)
        sys.exit(1)

class _MethodProcessor(object):

    def __init__(self, app, base_method_spec):
        self.app = app
        self.base_method_spec = base_method_spec
        self._method_names_by_order = []
        self._methods_by_name = {}
        self._build_method_index()
        self._method_spec_map = self._get_method_spec_map()

    def get_method_names_by_order(self):
        return list(self._method_names_by_order)

    def get_methods_by_name(self):
        return dict(self._methods_by_name)

    def get_method_spec_map(self):
        return dict(self._method_spec_map)

    def resolve_method_name(self, raw_method_name, default_method):
        method_name = raw_method_name if raw_method_name is not None else default_method
        if method_name is None:
            return None
        method_name = _to_snake(method_name.lstrip('-'))
        if method_name in self._method_spec_map:
            return method_name
        return None

    def _get_method_spec_map(self):
        method_spec_map = {}
        for method_name in self._method_names_by_order:
            method_spec_map[method_name] = self._get_method_spec(method_name)
        return method_spec_map

    def _get_method_spec(self, method_name):
        method_attr = self._methods_by_name[method_name]
        return _MethodSpec(method_attr, self.base_method_spec)

    def build_help_for_method(self, method_name):
        method_spec = self._method_spec_map.get(method_name)
        if not method_spec:
            return ''
        return method_spec.build_help_text()

    def _append_methods_with_order_override(self, new_methods):
        for method_name, method_attr in new_methods.items():
            if method_name in self._methods_by_name:
                self._method_names_by_order.remove(method_name)
            self._method_names_by_order.append(method_name)
            self._methods_by_name[method_name] = method_attr

    def _build_method_index(self):
        # 每次重建都先清空：避免同一个 AppSpec 实例被重复初始化时累计旧状态，保证顺序与映射始终从当前 app 重新计算。
        self._method_names_by_order = []
        self._methods_by_name = {}
        # 组合模式：__proxy_src_cls__ 是当前 App 本体，__proxy_cls_list__ 是 merge_class_for_composition 注入的扩展类列表。
        # 处理顺序固定为“本体 -> 扩展类列表”，同名方法按 _append_methods_with_order_override 规则后者覆盖并后移到末尾。
        if '__proxy_cls_list__' in self.app.__dict__:
            proxy_src_cls = self.app.__dict__['__proxy_src_cls__']
            proxy_cls_list = self.app.__dict__['__proxy_cls_list__']
            self._append_methods_with_order_override(self._get_cls_cli_methods(proxy_src_cls, True))
            for cls in proxy_cls_list:
                self._append_methods_with_order_override(self._get_cls_cli_methods(cls, False))
        else:
            # 普通模式：仅扫描当前 app 类的 CLI 可见方法（按继承过滤规则），无需处理组合覆盖链。
            self._append_methods_with_order_override(self._get_cls_cli_methods(self.app, True))

    def _get_cls_cli_methods(self, cls, exclude_parent=True):
        methods = self._get_cls_public_methods(cls)
        if exclude_parent:
            parent_methods = self._get_cls_parent_methods_for_exclude(cls)
            return {key: value for key, value in methods.items() if key not in parent_methods}
        return methods

    def _get_cls_public_methods(self, cls):
        members = inspect.getmembers(cls, predicate=lambda x: inspect.isfunction(x) or inspect.ismethod(x))
        return {item[0]: item[1] for item in members if item[0][0:1] != '_'}

    def _get_cls_parent_methods_for_exclude(self, cls):
        methods = {}
        for parent_cls in cls.__bases__:
            if parent_cls.__dict__.get('keep_class_for_cli', False):
                continue
            parent_methods = self._get_cls_public_methods(parent_cls)
            parent_methods = {
                key: val for key, val in parent_methods.items()
                if not val.__dict__.get('keep_method_for_cli', False)
            }
            methods.update(parent_methods)
        return methods

def _convert_value_for_type(value, type):
    default = False if type == 'bool' else None
    strict = type != 'bool'
    return datatypes.convert_scalar_value(value, type, default=default, strict=strict)

def _to_kebab(name):
    return name.replace('_', '-')

def _to_snake(name):
    return name.replace('-', '_')

def _build_main_long_option(name):
    return '--' + _to_kebab(name)

def _build_compat_long_option(name):
    if '_' in name:
        return '--' + name
    return None

def _build_method_help_prog(method_name):
    script_name = Path(sys.argv[0]).name
    return f'{script_name} {_to_kebab(method_name)}'

def _token_has_long_option(token, option):
    return token == option or token.startswith(option + '=')

def _check_mixed_long_option_styles(canonical_names, tokens):
    for name in canonical_names:
        compat_option = _build_compat_long_option(name)
        if compat_option is None:
            continue
        main_option = _build_main_long_option(name)
        has_main = any(_token_has_long_option(token, main_option) for token in tokens)
        has_compat = any(_token_has_long_option(token, compat_option) for token in tokens)
        if has_main and has_compat:
            raise ValueError('mixed option styles for %s: use either %s or %s' % (name, main_option, compat_option))

def _add_option_for_argument(parser, arg_name, metavar, help_text, has_default=False, default=None):
    if len(arg_name) == 1:
        parser.add_argument('-' + arg_name, metavar=metavar, help=help_text, default=default, nargs=1, required=not has_default)
        return
    main_option = _build_main_long_option(arg_name)
    compat_option = _build_compat_long_option(arg_name)
    if compat_option is None:
        parser.add_argument(main_option, dest=arg_name, metavar=metavar, help=help_text, default=default, nargs=1, required=not has_default)
        return
    if not has_default:
        group = parser.add_mutually_exclusive_group(required=True)
        group.add_argument(main_option, dest=arg_name, metavar=metavar, help=help_text, nargs=1)
        group.add_argument(compat_option, dest=arg_name, metavar=metavar, help=argparse.SUPPRESS, nargs=1)
        return
    parser.add_argument(main_option, dest=arg_name, metavar=metavar, help=help_text, default=default, nargs=1, required=False)
    parser.add_argument(compat_option, dest=arg_name, metavar=metavar, help=argparse.SUPPRESS, default=default, nargs=1, required=False)

def _add_method_placeholder_options(parser, method_name):
    # 兼容解析层：注册并隐藏 --cmd 占位，让 legacy 方法选项被 argparse 消费而不进入 remainder。
    main_option = _build_main_long_option(method_name)
    parser.add_argument(main_option, dest=method_name, action='store_true', help=argparse.SUPPRESS)
    compat_option = _build_compat_long_option(method_name)
    if compat_option is None:
        return
    parser.add_argument(compat_option, dest=method_name, action='store_true', help=argparse.SUPPRESS)

def build_argparser_for_func(method_name, method_spec):
    """根据 method_spec 构建 argparse，解析命令行，返回 kwargs（含 varargs 的 list）。

    参数映射：
    - 具名参数：单字符用 -x，多字符用 --name；有默认值则 optional，否则 required。
    - varargs：不建选项，由 parse_known_args 的 unknown 获取 remainder，作为 list 传入。
    - 子命令兼容占位：添加 --method-name（兼容 --method_name）仅用于兼容解析，不作为主调用方式展示。
    """
    args = list(method_spec.named_args)
    default_values = method_spec.default_values
    varargs_name = method_spec.varargs
    # varargs 不建 --option，从 remainder 获取
    if varargs_name:
        args = [a for a in args if a != varargs_name]

    parser = argparse.ArgumentParser(prog=_build_method_help_prog(method_name))
    # 子命令兼容占位，便于 parse_known_args 识别 legacy --cmd 输入
    _add_method_placeholder_options(parser, method_name)

    # 含 des、type 等，用于 help 与类型转换
    parameters = method_spec.parameters
    check_names = [method_name]
    for arg_name in args:
        parameter = parameters.get(arg_name)
        help = parameter.get('des', '')
        metavar = ''
        # 单字符用 -x；多字符主写法 --arg-name，兼容 --arg_name；nargs=1 统一取单值
        has_default = default_values is not None and arg_name in default_values
        default = default_values[arg_name] if has_default else None
        _add_option_for_argument(parser, arg_name, metavar, help, has_default=has_default, default=default)
        if len(arg_name) > 1:
            check_names.append(arg_name)
    _check_mixed_long_option_styles(check_names, sys.argv[1:])
    # unknown 为未被识别的 token，即 remainder（含 varargs）
    args_input, unknown = parser.parse_known_args()
    kwargs = {}
    for arg_name in args:
        parameter = parameters.get(arg_name)
        attr = getattr(args_input, arg_name)
        raw_value = _get_first_from_list_or_item_self(attr)
        kwargs[arg_name] = _convert_value_for_type(raw_value, parameter['type'])
    if varargs_name:
        remainder = list(unknown)
        # 子命令以词形式传入（如 manager run_tests）时会进 unknown，以选项传入（manager --run_tests）则被消费；若在首位则剔除。
        if remainder and _to_snake(remainder[0].lstrip('-')) == method_name:
            remainder = remainder[1:]
        # 剩余项作为 *args 的 list 传入
        kwargs[varargs_name] = remainder
    return kwargs

def _get_first_from_list_or_item_self(attr):
    # nargs=1 gives list (unwrap to scalar); with default and no arg, argparse may return scalar as-is.
    if isinstance(attr, (list, tuple)):
        return attr[0] if len(attr) > 0 else None
    return attr

def run_app(app, default_method=None, **kwargs):
    if 'base_args_option' in kwargs:
        raise ValueError('base_args_option is removed; put common args in app __init__ signature')
    app_spec = AppSpec(app)
    app_spec.resolve(default_method)
    app_spec.run()



import logging
import logging.handlers
import os
import sys
from datetime import datetime
from threading import get_ident

logger = logging.getLogger(__name__)

basic_msg_pre = '%(time_iso_8601)s %(local_ip)s %(pid)d.%(tid)d '
full_format_str = basic_msg_pre + '%(levelname)s %(name)s@%(filename)s:%(lineno)d %(message)s'

short_format_str = '%(time_short)s %(levelname)s %(name)s@%(filename_short)s:%(lineno)d %(message)s'


class Colors:
    RESET = '\033[0m'
    BOLD = '\033[1m'
    DEBUG = '\033[36m'
    INFO = '\033[32m'
    WARNING = '\033[33m'
    ERROR = '\033[31m'
    CRITICAL = '\033[35m'


def _supports_color():
    if not hasattr(sys.stdout, 'isatty') or not sys.stdout.isatty():
        return False
    if sys.platform == 'win32':
        try:
            import ctypes
            kernel32 = ctypes.windll.kernel32
            kernel32.SetConsoleMode(kernel32.GetStdHandle(-11), 7)
            return True
        except Exception:
            return False
    term = os.environ.get('TERM', '')
    if term and term != 'dumb':
        return True
    return False


def get_local_ip_address():
    try:
        import socket
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        s.connect(('8.8.8.8', 80))
        ip = s.getsockname()[0]
        s.close()
        if ip:
            return ip
    except Exception:
        pass
    return '127.0.0.1'


_level_abbrev = {
    'DEBUG': 'DBUG',
    'INFO': 'INFO',
    'WARNING': 'WARN',
    'ERROR': 'ERR ',
    'CRITICAL': 'CRIT',
}


class AlignedFormatter(logging.Formatter):

    def format(self, record):
        record.levelname = _level_abbrev.get(record.levelname, record.levelname)
        return super().format(record)


class ColoredFormatter(logging.Formatter):

    def __init__(self, format_str):
        super().__init__(format_str)
        self.color_map = {
            logging.DEBUG: Colors.DEBUG,
            logging.INFO: Colors.INFO,
            logging.WARNING: Colors.WARNING,
            logging.ERROR: Colors.ERROR,
            logging.CRITICAL: Colors.BOLD + Colors.CRITICAL,
        }

    def format(self, record):
        level_color = self.color_map.get(record.levelno, '')
        reset = Colors.RESET
        levelname_abbrev = _level_abbrev.get(record.levelname, record.levelname)
        record.levelname = f'{level_color}{levelname_abbrev}{reset}'
        return super().format(record)


class LogManager:

    def __init__(self):
        self.basic_handlers = []
        self.basic_log_level = logging.NOTSET
        self.disable_stdout_logger = False
        self.disable_file_logger = False
        self._init_from_env_done = False
        self._filelog_init_keys = set()

    def shutdown(self):
        for handler in self.basic_handlers:
            handler.close()

    def init_from_env(self, env_config):
        if self._init_from_env_done:
            return
        explicit_level = env_config.get('LOG.LEVEL')
        if explicit_level is not None:
            level = explicit_level
        else:
            app_env = (env_config.get('APP_ENV') or 'prod').lower()
            if app_env in ('dev', 'test'):
                level = 'debug'
            else:
                level = 'info'

        self.silence_noise_loggers(env_config)
        self.basic_log_level = _map_log_level(level)
        self.disable_stdout_logger = env_config.get_bool('LOG.DISABLE_STDOUT_LOGGER', default=False)
        if not self.disable_stdout_logger:
            self._init_stdout_logger()
        self._init_from_env_done = True
        logger.debug(f'root logger: {logging.getLogger()}, handlers: {self.basic_handlers}')

    def setup_readable_logger_handler(self, handler, use_color=False, use_short_format=False):
        format_str = short_format_str if use_short_format else full_format_str
        if use_color and _supports_color():
            formatter = ColoredFormatter(format_str)
        else:
            formatter = AlignedFormatter(format_str)
        filter_obj = BasicContextFilter(use_short_format=use_short_format)
        handler.setFormatter(formatter)
        handler.setLevel(self.basic_log_level)
        handler.addFilter(filter_obj)
        return handler

    def silence_noise_loggers(self, env_config):
        names = env_config.get_list('LOG.NOISE_LOGGERS')
        level = env_config.get('LOG.NOISE_LEVEL', 'warning')
        log_level = _map_log_level(level)
        for name in names:
            logging.getLogger(name).setLevel(log_level)

    def reset_root_logger_handlers(self):
        root_logger = logging.getLogger()
        root_logger.setLevel(self.basic_log_level)
        root_logger.handlers = []
        for handler in self.basic_handlers:
            root_logger.addHandler(handler)

    def _init_stdout_logger(self):
        stream_handler = logging.StreamHandler(sys.stdout)
        use_color = True
        use_short_format = True
        self.setup_readable_logger_handler(stream_handler, use_color=use_color, use_short_format=use_short_format)
        self.basic_handlers.append(stream_handler)
        self.reset_root_logger_handlers()

    def add_file_logger(self, app_name, logs_dir, use_short_format=False):
        key = (app_name, str(logs_dir), use_short_format)
        if key in self._filelog_init_keys:
            return
        filename = f'{logs_dir}/{app_name}.log'
        app_filelog_handler = _get_rotating_filehandler(filename)
        logger.debug(f'add file logger: {app_name}, {filename}, {use_short_format}')
        self.setup_readable_logger_handler(app_filelog_handler, use_short_format=use_short_format)
        self.basic_handlers.append(app_filelog_handler)
        self._filelog_init_keys.add(key)
        self.reset_root_logger_handlers()


log_manager = LogManager()


class BasicContextFilter(logging.Filter):

    def __init__(self, use_short_format=False):
        self.local_ip = get_local_ip_address()
        self.use_short_format = use_short_format

    def filter(self, record):
        if self.use_short_format:
            record.time_short = datetime.now().strftime('%H:%M:%S')
            record.filename_short = os.path.basename(record.filename)
        else:
            record.time_iso_8601 = datetime.now().astimezone().strftime('%Y-%m-%dT%H:%M:%S%z')
            record.local_ip = self.local_ip
            record.pid = os.getpid()
            record.tid = get_ident()
        return True


def _get_rotating_filehandler(filename):
    os.makedirs(os.path.dirname(filename), exist_ok=True)
    handler = logging.handlers.RotatingFileHandler(filename, mode='a', maxBytes=1024 * 1024 * 128, backupCount=3)
    os.chmod(filename, 0o666)
    return handler


def _map_log_level(log_level_str):
    log_level = logging.INFO
    if log_level_str is not None:
        level = log_level_str.lower()
        log_level_conf = {
            'verbose': logging.NOTSET,
            'debug': logging.DEBUG,
            'info': logging.INFO,
            'warning': logging.WARNING,
            'warn': logging.WARNING,
            'error': logging.ERROR,
            'critical': logging.CRITICAL,
        }
        if level in log_level_conf:
            log_level = log_level_conf[level]
    return log_level



import os
import socket
import sys
from pathlib import Path

from cpbox.app.appconfig import appconfig
from cpbox.tool import file
from cpbox.tool import functocli
from cpbox.tool import spec
from cpbox.tool import template
from cpbox.tool import utils
from cpbox.app.tools import tibaselog

is_windows = os.name == 'nt'

class DevOpsAppConfig:

    def __init__(self, app_name, provider):
        appconfig.init(app_name, provider.get_env(), provider.get_app_config())


class DevOpsAppConfigContext:

    def __init__(self, app_name, kwargs):

        self.app_base_name = app_name
        self.kwargs = kwargs
        self.sandbox_version = kwargs.get('sandbox_version', 0)

        self.env = self._get_env()
        self.app_name = self._determine_app_name()

        self.construct_baisc_dir()

    def _determine_app_name(self):
        app_name = self.app_base_name
        if self.sandbox_version != 0:
            app_name = '%s-sandbox-%03d' % (app_name, self.sandbox_version)
        return app_name

    def construct_baisc_dir(self):
        default_data_dir = '/opt/data' if not is_windows else 'D:\\opt\\data'
        data_dir = Path(os.environ.get('CPBOX_DATA_DIR', default_data_dir))

        self.app_storage_dir = data_dir / self.app_name

        self.app_persistent_storage_dir = self.app_storage_dir / 'persistent'
        self.app_runtime_storage_dir = self.app_storage_dir / 'runtime'
        self.app_logs_dir = self.app_runtime_storage_dir / 'logs'

    def is_dev(self):
        return self.env == 'dev' or self.env == 'test'

    def _get_env(self):
        env = 'dev'
        if 'PUPPY_ENV' in os.environ:
            env = os.environ['PUPPY_ENV']
        if 'CPBOX_ENV' in os.environ:
            env = os.environ['CPBOX_ENV']
        if 'env' in self.kwargs:
            env = self.kwargs['env']
        return env

    def get_sandbox_code_dir(self):
        sandbox_code_dir = None
        base_dir = self.app_persistent_storage_dir
        if self.sandbox_version != 0:
            sandbox_code_dir = '%s/sandbox-code-%03d' % (base_dir, self.sandbox_version)
            sandbox_code_dir = Path(sandbox_code_dir)
        return sandbox_code_dir

    def get_root_dir(self):
        if hasattr(sys.modules['__main__'], '__file__'):
            root_dir = Path(sys.argv[0]).resolve().parent
            return root_dir
        else:
            root_dir = Path.cwd()
            return root_dir

    def ensure_dir_and_write_permission(self):
        file.ensure_dir(self.app_persistent_storage_dir)
        file.ensure_dir(self.app_runtime_storage_dir)
        file.ensure_dir(self.app_logs_dir)
        file.ensure_dir(self.app_logs_dir / 'syslog')


class DevOpsApp(functocli.BaseCliApp):

    def __init__(self, app_name, **kwargs):
        functocli.BaseCliApp.__init__(self)
        self.config_context = DevOpsAppConfigContext(app_name, kwargs)
        self.env = self.config_context.env

        self.construct_baisc_dir()

        self.logger = self._make_logger(app_name, kwargs)

        self.file_lock = None

    def _make_logger(self, app_name, kwargs):
        from cpbox.app.rootcontext import gcontext
        gcontext.try_enable_framework_mode()
        tibaselog.log_manager.add_file_logger(app_name, self.app_logs_dir)
        import logging
        return logging.getLogger(app_name)

    def construct_baisc_dir(self):

        context = self.config_context
        context.construct_baisc_dir()
        context.ensure_dir_and_write_permission()

        hostname = socket.gethostname()
        self.hostname_fqdn = hostname
        self.hostname_short = hostname.split('.', 1)[0]

        self.root_dir = context.get_sandbox_code_dir() if context.sandbox_version else context.get_root_dir()
        self._roles_dir = self.root_dir / 'roles'

        app_root_dir = self._roles_dir / context.app_base_name

        self.app_root_dir = app_root_dir
        self.app_config_dir = app_root_dir / 'config'
        self.app_templates_dir = app_root_dir / 'templates'
        self.app_scripts_dir = app_root_dir / 'scripts'

        self.app_storage_dir = context.app_storage_dir
        self.app_persistent_storage_dir = context.app_persistent_storage_dir
        self.app_runtime_storage_dir = context.app_runtime_storage_dir
        self.app_logs_dir = context.app_logs_dir

    def remove_container(self, name, force=False, dry_run=False):
        cmd = 'docker rm %s' % (name)
        if force:
            cmd = 'docker rm -f %s' % (name)
        self.shell_run(cmd, exit_on_error=False, dry_run=dry_run)

    def container_is_running(self, name):
        cmd = 'docker ps | grep %s' % (name)
        return self.run_cmd(cmd, log=False)[0] == 0

    def stop_container(self, name, timeout=300, dry_run=False):
        cmd = 'docker stop --time %d %s' % (timeout, name)
        self.shell_run(cmd, exit_on_error=False, dry_run=dry_run)

    def _check_lock(self):
        filepath = self.app_runtime_storage_dir / 'locks' / file.compute_lock_filepath(sys.argv)
        file_lock = file.obtain_lock(filepath)
        if file_lock is None:
            pid = 0
            with open(filepath, 'r') as f:
                pid = f.read()
            self.logger.warning('lock file exists, pid: %s => %s', pid, filepath)
            sys.exit(1)
        else:
            self.file_lock = file_lock

    def template_to(self, template_filename, dst, payload, app_name=None):
        template_payload = {'payload': payload}
        src = self.app_templates_dir_for(app_name) / template_filename
        template.render_to_file(src, template_payload, dst)
        self.logger.info('template_to: %s => %s', src, dst)

    def read_app_config(self, config_filebasename, app_name=None):
        if '.yml' not in config_filebasename:
            config_filebasename = config_filebasename + '.yml'
        data = utils.load_yaml(self.app_config_dir_for(app_name) / config_filebasename, {})
        return data

    def app_templates_dir_for(self, app_name=None):
        if app_name is None:
            return self.app_templates_dir
        return self._roles_dir / app_name / 'templates'

    def app_config_dir_for(self, app_name=None):
        if app_name is None:
            return self.app_config_dir
        return self._roles_dir / app_name / 'config'

    @functocli.keep_method_for_inheritance
    def check_health(self):
        print('%s is good' % (sys.argv[0]))

    @staticmethod
    def run_app(app, log_level='info', default_method=None, common_args_option=None):
        functocli.run_app(app, default_method=default_method)



import time
from datetime import datetime


class IDMan:

    def __init__(self):
        self.last_time = 0
        self.sequence = 0

    def next_id(self) -> int:
        now = int(time.time() * 1000)
        if now == self.last_time:
            self.sequence += 1
        else:
            self.last_time = now
            self.sequence = 0
        return (now << 4) | self.sequence


def explain_id(id_value: int) -> str:
    timestamp = id_value >> 4
    sequence = id_value & 0xF
    dt = datetime.fromtimestamp(timestamp / 1000)
    time_str = dt.strftime('%Y-%m-%d %H:%M:%S.%f')[:-3]
    return f'ID: {id_value} (时间: {time_str}, 序列号: {sequence})'


class IDTag:

    def __init__(self, get_tag_func):
        self.get_tag_func = get_tag_func

    def __str__(self):
        return f'{self.get_tag_func()}'


class IDClass:
    id_man = IDMan()

    def __init__(self, get_tag_func=None):
        self.id = self.id_man.next_id()
        self.base_tag = f'{self.__class__.__name__}/{self.id}'
        if hasattr(self, '_get_tag'):
            self.tag = IDTag(self._get_tag)
        else:
            self.tag = self.base_tag

    def explain_id(self):
        return explain_id(self.id)



import logging
import os
from pathlib import Path
import sys

from dotenv import dotenv_values

from cpbox.app.tools import tibaselog
from cpbox.tool import datatypes
from cpbox.tool import cls
from cpbox.tool import file

logger = logging.getLogger(__name__)


def _detect_start_dir():
    # 启动目录判定规则：
    # 1) 正常脚本方式启动（python xxx.py）时，优先使用入口脚本所在目录，后续向上查找 .env 会更符合项目预期。
    # 2) 交互式场景（如 REPL / notebook / python -c）通常没有 __main__.__file__，回退到当前工作目录 cwd。
    if hasattr(sys.modules['__main__'], '__file__'):
        return Path(sys.argv[0]).resolve().parent
    return Path.cwd()


def detect_framework_root_dir(start_dir=None):
    if start_dir is None:
        start_dir = _detect_start_dir()
    else:
        start_dir = Path(start_dir).resolve()
    while True:
        if (start_dir / '.env').exists():
            return start_dir
        if start_dir == start_dir.parent:
            return None
        start_dir = start_dir.parent


def load_env_config_by_priority(framework_root_dir=None):
    if framework_root_dir is None:
        return EnvConfig(dict(os.environ))
    env_dict = dotenv_values(framework_root_dir / '.env')
    if (framework_root_dir / '.env.local').exists():
        env_local_dict = dotenv_values(framework_root_dir / '.env.local')
        env_dict.update(env_local_dict)
    for key, value in env_dict.items():
        if value is None:
            continue
        if key in os.environ:
            continue
        os.environ[key] = value
    return EnvConfig(dict(os.environ))


class EnvConfig(dict):

    def __init__(self, data):
        dict.__init__(self, data or {})

    def _to_type(self, raw, item_type, default, strict):
        return datatypes.convert_scalar_value(raw, item_type, default=default, strict=strict)

    def get_bool(self, key, default=False):
        return self._to_type(self.get(key, None), bool, default, strict=False)

    def get_int(self, key, default=0, strict=True):
        return self._to_type(self.get(key, None), int, default, strict)

    def get_float(self, key, default=0.0, strict=True):
        return self._to_type(self.get(key, None), float, default, strict)

    def get_list(self, key, item_type=str, sep=',', default=None):
        raw = self.get(key, None)
        if raw is None:
            return default if default is not None else []
        items = str(raw).split(sep)
        items = [item.strip() for item in items]
        items = [item for item in items if item]
        return [self._to_type(item, item_type, default, strict=True) for item in items]


def _normalize_app_env(env_config):
    app_env = env_config.get('APP_ENV') or env_config.get('ENV') or 'prod'
    app_env = str(app_env).lower()
    if app_env not in ('dev', 'test', 'pre', 'prod'):
        return 'prod'
    return app_env


class GContext(cls.Singleton):

    def __init__(self):
        super().__init__()

    def _init_once(self):
        self.mode = 'basic'
        self.framework_root_dir = None
        self.output_dir = None
        self.env_config = EnvConfig(dict())
        self.is_dev_mode = False
        self._has_init_done = False

    def try_enable_framework_mode(self):
        if self._has_init_done:
            return
        start_dir = _detect_start_dir()
        self.framework_root_dir = detect_framework_root_dir(start_dir)
        self.mode = 'framework' if self.framework_root_dir is not None else 'basic'
        self.output_dir = None
        self.env_config = load_env_config_by_priority(self.framework_root_dir)
        app_env = _normalize_app_env(self.env_config)
        self.is_dev_mode = app_env in ('dev', 'test')
        self._init_logger()
        self._has_init_done = True

    def _init_logger(self):
        tibaselog.log_manager.init_from_env(self.env_config)
        if self.mode != 'framework':
            return
        self.output_dir = self.framework_root_dir / 'output'
        file.ensure_dir(self.output_dir)
        logs_dir = self.output_dir / 'logs'
        app_name = self.env_config.get('APP_NAME', 'tiapp')
        tibaselog.log_manager.add_file_logger(app_name, logs_dir)


gcontext = GContext()



import json
import os
from io import StringIO

from cpbox.tool import cls
from cpbox.tool import file


class YAMLInitializer(cls.Singleton):

    def _init_once(self):
        from ruamel.yaml import YAML
        self.yaml = YAML(typ='rt')
        self.yaml.indent(mapping=2, sequence=4, offset=2)
        self.yaml.allow_unicode = True
        self.yaml.sort_keys = False
        self.yaml.default_flow_style = False
        self.yaml.default_style = None


yaml = YAMLInitializer().yaml


def pyaml(data) -> str:
    stream = StringIO()
    yaml.dump(data, stream=stream)
    return stream.getvalue()


def pjson(data):
    return json.dumps(data, ensure_ascii=False, indent=2)


def load_yaml_file(fn, fallback=None):
    if not os.path.isfile(fn):
        return fallback
    with open(fn, 'r', encoding='utf-8') as input_file:
        data = yaml.load(input_file)
    return data if data is not None else fallback


def dump_yaml_file(fn, data):
    file.ensure_dir_for_file(str(fn))
    with open(fn, 'w', encoding='utf-8') as outfile:
        yaml.dump(data, outfile)


def load_json_file(fn, fallback=None):
    if not os.path.isfile(fn):
        return fallback
    try:
        with open(fn, 'r', encoding='utf-8') as input_file:
            data = json.load(input_file)
        return data if data is not None else fallback
    except Exception:
        return fallback


def dump_json_file(fn, data):
    file.ensure_dir_for_file(str(fn))
    with open(fn, 'w', encoding='utf-8') as output_file:
        json.dump(data, output_file, ensure_ascii=False, indent=2)



from typing import Any
from typing import Dict
from typing import Type


class Singleton:
    _instances: Dict[Type, Any] = {}

    def __new__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super().__new__(cls)
        return cls._instances[cls]

    def __init__(self):
        if not hasattr(self, '_initialized'):
            self._initialized = True
            self._init_once()

    def _init_once(self):
        pass
